@@ -1,7 +1,10 @@
+# -*- coding: utf-8
 """
 File containing ConfigObject.
 """
 
+from __future__ import division, print_function, unicode_literals
+
 ### Logging ###
 import logging
 _logger = logging.getLogger("ConfigUtils")
@@ -9,9 +12,12 @@ _logger = logging.getLogger("ConfigUtils")
 
 import os
 import sys
-import ConfigParser as configparser
 from ast import literal_eval
-from gettext import gettext as _
+try:
+    import configparser
+except ImportError:
+    # python2 fallback
+    import ConfigParser as configparser
 
 from gi.repository import Gio
 
@@ -33,11 +39,11 @@ class ConfigObject(object):
     Python properties and notification functions are created
     automagically for all keys added in _init_keys().
     """
-    def __init__(self, parent = None, gspath = ""):
+    def __init__(self, parent = None, schema = ""):
         self.parent = parent       # parent ConfigObject
         self.children = []         # child config objects; not necessarily
                                    #   reflecting the gsettings hierarchy
-        self.gspath = gspath       # path to the gsettings object
+        self.schema = schema       # schema-path to the gsettings object
         self.gskeys = {}           # key-value objects {property name, GSKey()}
         self.sysdef_section = None # system defaults section name
         self.system_defaults = {}  # system defaults {property name, value}
@@ -46,13 +52,13 @@ class ConfigObject(object):
         self._init_keys()
 
         # check if the gsettings schema is installed
-        if not self.gspath in Gio.Settings.list_schemas():
+        if not self.schema in Gio.Settings.list_schemas():
             raise SchemaError(_("gsettings schema for '{}' is not installed").
-                                                             format(self.gspath))
+                                                             format(self.schema))
 
         # create gsettings object and its python properties
-        self.settings = Gio.Settings.new(self.gspath)
-        for gskey in self.gskeys.values():
+        self.settings = Gio.Settings.new(self.schema)
+        for gskey in list(self.gskeys.values()):
             gskey.settings = self.settings
             self._setup_property(gskey)
 
@@ -63,12 +69,28 @@ class ConfigObject(object):
         """ overload this and use add_key() to add key-value tuples """
         pass
 
-    def add_key(self, key, default, prop = None, sysdef = None):
+    def add_key(self, key, default, prop = None, sysdef = None, 
+                      writable = True):
         """ Convenience function to create and add a new GSKey. """
-        gskey = GSKey(None, key, default, prop, sysdef)
+        gskey = GSKey(None, key, default, prop, sysdef, writable)
         self.gskeys[gskey.prop] = gskey
         return gskey
 
+    def find_key(self, key):
+        """ Search for key (gsettings name) """
+        for gskey in self.gskeys.values():
+            if gskey.key == key:
+                return gskey
+        return None
+
+    def get_root(self):
+        """ Return the root config object """
+        co = self
+        while co:
+            if co.parent is None:
+                return co
+            co = co.parent
+
     def check_hooks(self):
         """
         Simple runtime plausibility check for all overloaded hook functions.
@@ -93,7 +115,7 @@ class ConfigObject(object):
 
     def disconnect_notifications(self):
         """ Recursively remove all callbacks from all notification lists. """
-        for gskey in self.gskeys.values():
+        for gskey in list(self.gskeys.values()):
             prop = gskey.prop
             setattr(type(self), _NOTIFY_CALLBACKS.format(prop), [])
 
@@ -107,12 +129,21 @@ class ConfigObject(object):
         # list of callbacks
         setattr(type(self), _NOTIFY_CALLBACKS.format(prop), [])
 
-        # method to add callbak
+        # method to add callback
         def _notify_add(self, callback, _prop=prop):
             """ method to add a callback to this property """
             getattr(self, _NOTIFY_CALLBACKS.format(prop)).append(callback)
         setattr(type(self), prop+'_notify_add', _notify_add)
 
+        # method to remove a callback
+        def _notify_remove(self, callback, _prop=prop):
+            """ method to remove a callback from this property """
+            try:
+                getattr(self, _NOTIFY_CALLBACKS.format(prop)).remove(callback)
+            except ValueError:
+                pass
+        setattr(type(self), prop+'_notify_remove', _notify_remove)
+
         # gsettings callback
         def _notify_changed_cb(self, settings, key, _gskey=gskey, _prop=prop):
             """ call back function for change notification """
@@ -130,8 +161,18 @@ class ConfigObject(object):
                 if _gskey.value != value:
                     _gskey.value = value
 
-                    for callback in getattr(self, _NOTIFY_CALLBACKS.format(prop)):
-                        callback(value)
+                    if False:
+                        # asynchronous callbacks
+                        def notify(callbacks, value):
+                            for callback in callbacks:
+                                callback(value)
+
+                        GObject.idle_add(notify, 
+                                        getattr(self, _NOTIFY_CALLBACKS.format(prop)),
+                                        value)
+                    else:
+                        for callback in getattr(self, _NOTIFY_CALLBACKS.format(prop)):
+                            callback(value)
 
             # Post-notification hook for anything that properties
             # need to do after all listeners have been notified.
@@ -190,16 +231,16 @@ class ConfigObject(object):
             self.use_system_defaults = False    # write to gsettings
 
         # let command line options override everything
-        for gskey in self.gskeys.values():
-            if hasattr(options, gskey.key):  # command line option there?
-                value = getattr(options, gskey.key)
+        for gskey in list(self.gskeys.values()):
+            if hasattr(options, gskey.prop):  # command line option there?
+                value = getattr(options, gskey.prop)
                 if not value is None:
                     gskey.value = value
 
     def init_from_gsettings(self):
         """ init propertiy values from gsettings """
 
-        for prop, gskey in self.gskeys.items():
+        for prop, gskey in list(self.gskeys.items()):
             gskey.value = gskey.default
             if hasattr(self, _GSETTINGS_GET_HOOK + prop):
                 gskey.value = getattr(self, _GSETTINGS_GET_HOOK + prop)(gskey)
@@ -212,7 +253,7 @@ class ConfigObject(object):
     def init_from_system_defaults(self):
         """ fill property values with system defaults """
 
-        for prop, value in self.system_defaults.items():
+        for prop, value in list(self.system_defaults.items()):
             setattr(self, prop, value)  # write to gsettings
 
         for child in self.children:
@@ -246,9 +287,9 @@ class ConfigObject(object):
         filepath = filename
         if filename and not os.path.exists(filename):
             # assume filename is just a basename instead of a full file path
-            _logger.debug(_("%s '%s' not found yet, "
-                           "retrying in default paths") %
-                         (description, filename))
+            _logger.debug(_("{description} '{filename}' not found yet, "
+                           "retrying in default paths") \
+                           .format(description=description, filename=filename))
 
             if user_filename_func:
                 filepath = user_filename_func(filename)
@@ -261,35 +302,76 @@ class ConfigObject(object):
                     filepath = ""
 
             if not filepath:
-                _logger.info(_("unable to locate '%s', "
-                               "loading default %s instead") %
-                             (filename, description))
+                _logger.info(_("unable to locate '{filename}', "
+                               "loading default {description} instead") \
+                            .format(description=description, filename=filename))
         if not filepath and not final_fallback is None:
             filepath = final_fallback
 
         if not os.path.exists(filepath):
-            _logger.error(_("failed to find %s '%s'") % (description, filename))
+            _logger.error(_("failed to find {description} '{filename}'") \
+                           .format(description=description, filename=filename))
             filepath = ""
         else:
-            _logger.debug(_("{} '{}' found.").format(description, filepath))
+            _logger.debug(_("{description} '{filepath}' found.") \
+                          .format(description=description, filepath=filepath))
 
         return filepath
 
     @staticmethod
-    def _dict_to_gsettings_list(gskey, _dict):
+    def get_unpacked_string_list(gskey, type_spec):
         """ Store dictionary in a gsettings list key """
-        _list = pack_name_value_list(_dict)
+        _list = gskey.settings.get_strv(gskey.key)
+        return ConfigObject.unpack_string_list(_list, type_spec)
+
+    @staticmethod
+    def set_packed_string_list(gskey, value):
+        """ Store dictionary in a gsettings list key """
+        _list = ConfigObject.pack_string_list(value)
         gskey.settings.set_strv(gskey.key, _list)
 
     @staticmethod
-    def _gsettings_list_to_dict(gskey, key_type = str):
+    def pack_string_list(value):
+        """ very crude hard coded behavior, fixme as needed """
+        if type(value) == dict:
+            _dict = value
+            if value:
+                # has collection interface?
+                key, _val = _dict.items()[0]
+                if not hasattr(_val, "__iter__"):
+                    _dict = dict([key, [value]] for key, value in _dict.items())
+            return ConfigObject._dict_to_list(_dict)
+
+        assert(False) # unsupported python type
+
+    @staticmethod
+    def unpack_string_list(_list, type_spec):
+        """ very crude hard coded behavior, fixme as needed """
+        if type_spec == "a{ss}":
+            _dict = ConfigObject._list_to_dict(_list, str, num_values = 1)
+            return dict([key, value[0]] for key, value in _dict.items())  
+
+        if type_spec == "a{s[ss]}":
+            return ConfigObject._list_to_dict(_list, str, num_values = 2)
+
+        if type_spec == "a{i[ss]}":
+            return ConfigObject._list_to_dict(_list, int, num_values = 2)
+
+        assert(False) # unsupported type_spec
+
+    @staticmethod
+    def _dict_to_list(_dict):
+        """ Store dictionary in a gsettings list key """
+        return pack_name_value_list(_dict)
+
+    @staticmethod
+    def _list_to_dict(_list, key_type = str, num_values = 2):
         """ Get dictionary from a gsettings list key """
-        _list = gskey.settings.get_strv(gskey.key)
-
-        _list = [x.decode("utf-8") for x in _list]  # translate to unicode
-
-        return unpack_name_value_list(_list, key_type=key_type)
+        if sys.version_info.major == 2:
+            _list = [x.decode("utf-8") for x in _list]  # translate to unicode
 
+        return unpack_name_value_list(_list, key_type=key_type,
+                                             num_values = num_values)
 
     def load_system_defaults(self, paths):
         """
@@ -298,7 +380,8 @@ class ConfigObject(object):
         They are stored in simple ini-style files, residing in a small choice
         of directories. The last setting found in the list of paths wins.
         """
-        _logger.info(_("Looking for system defaults in %s") % str(paths))
+        _logger.info(_("Looking for system defaults in {paths}") \
+                        .format(paths=paths))
 
         filename = None
         parser = configparser.SafeConfigParser()
@@ -310,58 +393,63 @@ class ConfigObject(object):
         if not filename:
             _logger.info(_("No system defaults found."))
         else:
-            _logger.info(_("Loading system defaults from %s.") % filename)
-            self.read_sysdef_section(parser)
+            _logger.info(_("Loading system defaults from {filename}") \
+                            .format(filename=filename))
+            self._read_sysdef_section(parser)
 
 
-    def read_sysdef_section(self, parser):
+    def _read_sysdef_section(self, parser):
         """
         Read this instances (and its childrens) system defaults section.
         """
 
         for child in self.children:
-            child.read_sysdef_section(parser)
+            child._read_sysdef_section(parser)
 
         self.system_defaults = {}
         if self.sysdef_section and \
            parser.has_section(self.sysdef_section):
             items = parser.items(self.sysdef_section)
-            items = [(key, val.decode("UTF-8")) for key, val in items]
+
+            if sys.version_info.major == 2:
+                items = [(key, val.decode("UTF-8")) for key, val in items]
 
             # convert ini file strings to property values
-            sysdef_gskeys = dict((k.sysdef, k) for k in self.gskeys.values())
+            sysdef_gskeys = dict((k.sysdef, k) for k in list(self.gskeys.values()))
             for sysdef, value in items:
-                _logger.info(_(u"Found system default '{}={}'") \
-                              .format(sysdef, value))
+                _logger.info(_("Found system default '[{}] {}={}'") \
+                              .format(self.sysdef_section, sysdef, value))
 
                 gskey = sysdef_gskeys.get(sysdef, None)
-                value = self.convert_sysdef_key(gskey, sysdef, value)
+                value = self._convert_sysdef_key(gskey, sysdef, value)
 
                 if not value is None:
                     prop = gskey.prop if gskey else sysdef.replace("-", "_")
                     self.system_defaults[prop] = value
 
 
-    def convert_sysdef_key(self, gskey, sysdef, value):
+    def _convert_sysdef_key(self, gskey, sysdef, value):
         """
         Convert a system default string to a property value.
         Sysdef strings -> values of type of gskey's default value.
         """
 
         if gskey is None:
-            _logger.warning(_(u"System defaults: Unknown key '{}' "
-                              u"in section '{}'") \
+            _logger.warning(_("System defaults: Unknown key '{}' "
+                              "in section '{}'") \
                               .format(sysdef, self.sysdef_section))
         else:
             _type = type(gskey.default)
-            if _type == str and value[0] != u'"':
-                value = u'"' + value + '"'
+            str_type = str if sys.version_info.major >= 3 \
+                       else unicode
+            if _type == str_type and value[0] != '"':
+                value = '"' + value + '"'
             try:
                 value = literal_eval(value)
             except (ValueError, SyntaxError) as ex:
-                _logger.warning(_(u"System defaults: Invalid value"
-                                  u" for key '{}' in section '{}'"
-                                  u"\n  {}").format(sysdef,
+                _logger.warning(_("System defaults: Invalid value"
+                                  " for key '{}' in section '{}'"
+                                  "\n  {}").format(sysdef,
                                                     self.sysdef_section, ex))
                 return None  # skip key
         return value
@@ -373,17 +461,22 @@ class GSKey:
     It associates python properties with gsettings keys,
     system default keys and command line options.
     """
-    def __init__(self, settings, key, default, prop, sysdef):
+    def __init__(self, settings, key, default, prop, sysdef, writable):
         if prop is None:
             prop = key.replace("-","_")
         if sysdef is None:
             sysdef = key
-        self.settings = settings # gsettings object
-        self.key      = key      # gsettings key name
-        self.sysdef   = sysdef   # system default name
-        self.prop     = prop     # python property name
-        self.default  = default  # hard coded default, determines type
-        self.value    = default  # current property value
+        self.settings  = settings # gsettings object
+        self.key       = key      # gsettings key name
+        self.sysdef    = sysdef   # system default name
+        self.prop      = prop     # python property name
+        self.default   = default  # hard coded default, determines type
+        self.value     = default  # current property value
+        self.writable = writable  # If False, never write the key to gsettings
+                                  #    even on accident.
+
+    def is_default(self):
+        return self.value == self.default
 
     def gsettings_get(self):
         """ Get value from gsettings. """
@@ -391,7 +484,7 @@ class GSKey:
         try:
             # Bug in Gio, gir1.2-glib-2.0, Oneiric
             # Onboard is accumultating open file handles
-            # of "/home/<user>/.config/dconf/<user>' when
+            # at "/home/<user>/.config/dconf/<user>' when
             # reading from gsettings before writing.
             # Check with:
             # lsof -w -p $( pgrep gio-test ) -Fn |sort|uniq -c|sort -n|tail
@@ -414,9 +507,11 @@ class GSKey:
 
     def gsettings_set(self, value):
         """ Send value to gsettings. """
-        self.settings[self.key] = value
+        if self.writable:
+            self.settings[self.key] = value
 
     def gsettings_apply(self):
         """ Send current value to gsettings. """
-        self.settings[self.key] = self.value
+        if self.writable:
+            self.settings[self.key] = self.value
 
