@@ -1,7 +1,6 @@
-#!/usr/bin/python
 # -*- coding: utf-8 -*-
 
-from __future__ import with_statement
+from __future__ import division, print_function, unicode_literals
 
 ### Logging ###
 import logging
@@ -10,10 +9,8 @@ _logger = logging.getLogger("KeyboardSVG")
 
 import os
 import re
-import string
 import sys
 import shutil
-from gettext import gettext as _
 from xml.dom import minidom
 
 from Onboard             import Exceptions
@@ -23,7 +20,8 @@ from Onboard.Keyboard    import Keyboard
 from Onboard.KeyboardGTK import KeyboardGTK
 from Onboard.Layout      import LayoutBox, LayoutPanel
 from Onboard.Appearance  import ColorScheme
-from Onboard.utils       import hexstring_to_float, modifiers, Rect, toprettyxml
+from Onboard.utils       import hexstring_to_float, modifiers, Rect, \
+                                toprettyxml, Version
 
 ### Config Singleton ###
 from Onboard.Config import Config
@@ -34,6 +32,15 @@ class KeyboardSVG(config.kbd_render_mixin, Keyboard):
     """
     Keyboard layout loaded from an SVG file.
     """
+    # onboard <= 0.95
+    LAYOUT_FORMAT_LEGACY      = Version(1, 0)
+
+    # onboard 0.96, initial layout-tree
+    LAYOUT_FORMAT_LAYOUT_TREE = Version(2, 0)
+
+    # onboard 0.97, scanner overhaul, no more scan columns, 
+    # new attributes scannable, scan_priority
+    LAYOUT_FORMAT             = Version(2, 1)  
 
     def __init__(self, vk, layout_filename, color_scheme_filename):
         config.kbd_render_mixin.__init__(self)
@@ -45,6 +52,10 @@ class KeyboardSVG(config.kbd_render_mixin, Keyboard):
 
         self.initial_update()
 
+    def initial_update(self):
+        config.kbd_render_mixin.initial_update(self)
+        Keyboard.initial_update(self)
+
     def destruct(self):
         config.kbd_render_mixin.destruct(self)
         Keyboard.destruct(self)
@@ -66,13 +77,16 @@ class KeyboardSVG(config.kbd_render_mixin, Keyboard):
             dom = minidom.parse(f).documentElement
 
             # check layout format
-            format = 1.0
+            format = self.LAYOUT_FORMAT_LEGACY
             if dom.hasAttribute("format"):
-               format = float(dom.attributes["format"].value)
+               format = Version.from_string(dom.attributes["format"].value)
 
-            if format >= 2.0:   # layout-tree format
+            if format >= self.LAYOUT_FORMAT_LAYOUT_TREE:
                 items = self._parse_dom_node(dom)
             else:
+                _logger.warning(_("Loading legacy layout format '{}'. "
+                            "Please consider upgrading to current format '{}'"
+                            ).format(format, self.LAYOUT_FORMAT))
                 items = self._parse_legacy_layout(dom)
 
             if items:
@@ -88,15 +102,13 @@ class KeyboardSVG(config.kbd_render_mixin, Keyboard):
         items = []
         for child in dom_node.childNodes:
             if child.nodeType == minidom.Node.ELEMENT_NODE:
-                if child.tagName == u"box":
+                if child.tagName == "box":
                     item = self._parse_box(child)
-                elif child.tagName == u"panel":
+                elif child.tagName == "panel":
                     item = self._parse_panel(child)
-                elif child.tagName == u"key":
+                elif child.tagName == "key":
                     item = self._parse_key(child, parent_item)
                 else:
-                    if child.tagName == u"column":    # scanning column
-                        self._parse_scan_column(child, parent_item)
                     item = None
 
                 if item:
@@ -120,6 +132,8 @@ class KeyboardSVG(config.kbd_render_mixin, Keyboard):
             item.visible = node.attributes["visible"].value == "true"
         if node.hasAttribute("border"):
             item.border = float(node.attributes["border"].value)
+        if node.hasAttribute("expand"):
+            item.expand = node.attributes["expand"].value == "true"
 
     def _parse_box(self, node):
         item = LayoutBox()
@@ -136,16 +150,6 @@ class KeyboardSVG(config.kbd_render_mixin, Keyboard):
         self._parse_dom_node_item(node, item)
         return item
 
-    def _parse_scan_column(self, node, parent):
-        column = []
-        for scanKey in node.getElementsByTagName("scankey"):
-            column.append(scanKey.attributes["id"].value)
-        columns = parent.scan_columns
-        if not columns:
-            columns = []
-        columns.append(column)
-        parent.scan_columns = columns
-
     def _parse_key(self, node, parent):
         key = RectKey()
         key.parent = parent # assign parent early to make get_filename() work
@@ -153,7 +157,7 @@ class KeyboardSVG(config.kbd_render_mixin, Keyboard):
         # parse standard layout item attributes
         self._parse_dom_node_item(node, key)
 
-        attributes = dict(node.attributes.items())
+        attributes = dict(list(node.attributes.items()))
         self._init_key(key, attributes)
 
         # get key geometry from the closest svg file
@@ -171,12 +175,7 @@ class KeyboardSVG(config.kbd_render_mixin, Keyboard):
                                       " Not found in '{}'.") \
                                     .format(key.theme_id, filename))
                 else:
-                    key.location = svg_key.location
-                    key.geometry = svg_key.geometry
-                    key.context.log_rect = Rect(svg_key.location[0],
-                                                svg_key.location[1],
-                                                svg_key.geometry[0],
-                                                svg_key.geometry[1])
+                    key.set_border_rect(svg_key.get_border_rect().copy())
                     return key
 
         return None  # ignore keys not found in an svg file
@@ -184,13 +183,7 @@ class KeyboardSVG(config.kbd_render_mixin, Keyboard):
     def _init_key(self, key, attributes):
         # Re-parse the id to distinguish between the short key_id
         # and the optional longer theme_id.
-        # The theme id has the form <id>.<arbitrary identifier>, where
-        # the identifier may be the name of the layout layer the key is
-        # defined in, e.g. 'DELE.compact-alpha'.
-        value = attributes["id"]
-        key.id = value.split(".")[0]
-        key.theme_id = value
-
+        key.set_id(attributes["id"])
 
         if "char" in attributes:
             key.action = attributes["char"]
@@ -199,16 +192,17 @@ class KeyboardSVG(config.kbd_render_mixin, Keyboard):
             value = attributes["keysym"]
             key.action_type = KeyCommon.KEYSYM_ACTION
             if value[1] == "x":#Deals for when keysym is hex
-                key.action = string.atoi(value,16)
+                key.action = int(value,16)
             else:
-                key.action = string.atoi(value,10)
+                key.action = int(value,10)
         elif "keypress_name" in attributes:
             key.action = attributes["keypress_name"]
             key.action_type = KeyCommon.KEYPRESS_NAME_ACTION
         elif "modifier" in attributes:
             try:
                 key.action = modifiers[attributes["modifier"]]
-            except KeyError, (strerror):
+            except KeyError as xxx_todo_changeme:
+                (strerror) = xxx_todo_changeme
                 raise Exception("Unrecognised modifier %s in" \
                     "definition of %s" (strerror, key.id))
             key.action_type = KeyCommon.MODIFIER_ACTION
@@ -220,8 +214,7 @@ class KeyboardSVG(config.kbd_render_mixin, Keyboard):
             key.action = attributes["script"]
             key.action_type = KeyCommon.SCRIPT_ACTION
         elif "keycode" in attributes:
-            key.action = string.atoi(
-                attributes["keycode"])
+            key.action = int(attributes["keycode"])
             key.action_type = KeyCommon.KEYCODE_ACTION
         elif "button" in attributes:
             key.action = key.id[:]
@@ -244,7 +237,7 @@ class KeyboardSVG(config.kbd_render_mixin, Keyboard):
         if "image" in attributes:
             key.image_filename = attributes["image"]
 
-        labels = [u"",u"",u"",u"",u""]
+        labels = ["","","","",""]
         #if label specified search for modified labels.
         if "label" in attributes:
             labels[0] = attributes["label"]
@@ -259,7 +252,7 @@ class KeyboardSVG(config.kbd_render_mixin, Keyboard):
                     attributes["altgrNshift_label"]
         # If key is a macro (snippet) generate label from number.
         elif key.action_type == KeyCommon.MACRO_ACTION:
-            label, text = config.snippets.get(string.atoi(key.action), \
+            label, text = config.snippets.get(int(key.action), \
                                                        (None, None))
             tooltip = _("Snippet {}").format(key.action)
             if not label:
@@ -269,7 +262,7 @@ class KeyboardSVG(config.kbd_render_mixin, Keyboard):
                 # Snippet n, unassigned - click to edit
                 tooltip += _(", unassigned")
             else:
-                labels[0] = label.replace(u"\\n", u"\n")
+                labels[0] = label.replace("\\n", "\n")
             key.tooltip = tooltip
 
         # Get labels from keyboard.
@@ -277,14 +270,15 @@ class KeyboardSVG(config.kbd_render_mixin, Keyboard):
             if key.action_type == KeyCommon.KEYCODE_ACTION:
                 if self.vk: # xkb keyboard found?
                     labDic = self.vk.labels_from_keycode(key.action)
-                    labDic = [x.decode("UTF-8") for x in labDic]
+                    if sys.version_info.major == 2:
+                        labDic = [x.decode("UTF-8") for x in labDic]
                     labels = (labDic[0],labDic[2],labDic[1],
                                             labDic[3],labDic[4])
                 else:
                     if key.id.upper() == "SPCE":
-                        labels = [u"No X keyboard found, retrying..."]*5
+                        labels = ["No X keyboard found, retrying..."]*5
                     else:
-                        labels = [u"?"]*5
+                        labels = ["?"]*5
 
         # Translate labels - Gettext behaves oddly when translating
         # empty strings
@@ -303,18 +297,6 @@ class KeyboardSVG(config.kbd_render_mixin, Keyboard):
 
         key.group = group_name
 
-        if "font_offset_x" in attributes:
-            offset_x = float(attributes["font_offset_x"])
-        else:
-            offset_x = config.DEFAULT_LABEL_OFFSET[0]
-
-        if "font_offset_y" in attributes:
-            offset_y = \
-                float(attributes["font_offset_y"])
-        else:
-            offset_y = config.DEFAULT_LABEL_OFFSET[1]
-        key.label_offset = (offset_x, offset_y)
-
         if "label_x_align" in attributes:
             key.label_x_align = float(attributes["label_x_align"])
         if "label_y_align" in attributes:
@@ -333,36 +315,17 @@ class KeyboardSVG(config.kbd_render_mixin, Keyboard):
         else:
             key.sticky = False
 
+        if "scannable" in attributes:
+            if attributes["scannable"].lower() == 'false':
+                key.scannable = False
+
+        if "scan_priority" in attributes:
+            key.scan_priority = int(attributes["scan_priority"])
+
         if "tooltip" in attributes:
             key.tooltip = attributes["tooltip"]
 
-        # old colors as fallback
-        rgba = [0.9, 0.85, 0.7]
-        key.rgba         = rgba
-        key.hover_rgba   = rgba
-        key.pressed_rgba = rgba
-        key.latched_rgba = [0.5, 0.5, 0.5, 1.0]
-        key.locked_rgba  = [1.0, 0.0, 0.0, 1.0]
-        key.scanned_rgba = [0.45, 0.45, 0.7, 1.0]
-        key.stroke_rgba  = [0.0, 0.0, 0.0, 1.0]
-        key.label_rgba   = [0.0, 0.0, 0.0, 1.0]
-
-        # get colors from color scheme
-        if self.color_scheme:
-            get_key_rgba = self.color_scheme.get_key_rgba
-            key.rgba                = get_key_rgba(key, "fill")
-            key.hover_rgba          = get_key_rgba(key, "hover")
-            key.pressed_rgba        = get_key_rgba(key, "pressed")
-            key.latched_rgba        = get_key_rgba(key, "latched")
-            key.locked_rgba         = get_key_rgba(key, "locked")
-            key.scanned_rgba        = get_key_rgba(key, "scanned")
-            key.stroke_rgba         = get_key_rgba(key, "stroke")
-            key.label_rgba          = get_key_rgba(key, "label")
-            key.dwell_progress_rgba = get_key_rgba(key, "dwell-progress")
-            key.color_scheme = self.color_scheme
-
-            is_key_default_color = self.color_scheme.is_key_default_color
-            key.pressed_rgba_is_default = is_key_default_color(key, "pressed")
+        key.color_scheme = self.color_scheme
 
 
 
@@ -381,7 +344,8 @@ class KeyboardSVG(config.kbd_render_mixin, Keyboard):
                 svg_dom = minidom.parse(svg_file).documentElement
                 svg_keys = self._parse_svg(svg_dom)
 
-        except Exception, (exception):
+        except Exception as xxx_todo_changeme1:
+            (exception) = xxx_todo_changeme1
             raise Exceptions.LayoutFileError(_("Error loading ")
                 + filename, chained_exception = exception)
 
@@ -392,15 +356,15 @@ class KeyboardSVG(config.kbd_render_mixin, Keyboard):
         for rect in svg_dom.getElementsByTagName("rect"):
             id = rect.attributes["id"].value
 
-            pos  = (float(rect.attributes['x'].value),
-                    float(rect.attributes['y'].value))
-            size = (float(rect.attributes['width'].value),
-                    float(rect.attributes['height'].value))
+            rect = Rect(float(rect.attributes['x'].value),
+                        float(rect.attributes['y'].value),
+                        float(rect.attributes['width'].value),
+                        float(rect.attributes['height'].value))
 
             # Use RectKey as cache for svg provided properties.
             # This key instance doesn't enter the layout and will
             # be discarded after the layout tree has been loaded.
-            key = RectKey(id, pos, size)
+            key = RectKey(id, rect)
 
             keys[id] = key
 
@@ -425,12 +389,22 @@ class KeyboardSVG(config.kbd_render_mixin, Keyboard):
             # parse keys
             keys = []
             for node in pane_node.getElementsByTagName("key"):
-                keys.append(self._parse_key(node, item))
+                key = self._parse_key(node, item)                
+                if key:
+                    # some keys have changed since Onboard 0.95
+                    if key.id == "middleClick":
+                        key.set_id("middleclick")
+                        key.action_type = KeyCommon.BUTTON_ACTION
+                    if key.id == "secondaryClick":
+                        key.set_id("secondaryclick")
+                        key.action_type = KeyCommon.BUTTON_ACTION
+                        
+                    keys.append(key)
+                    
             item.set_items(keys)
 
-            # parse scan columns
-            for node in pane_node.getElementsByTagName("column"):
-                self._parse_scan_column(node, item)
+            # check for scan columns
+            if pane_node.getElementsByTagName("column"):
                 is_scan = True
 
             panes.append(item)
@@ -444,7 +418,7 @@ class KeyboardSVG(config.kbd_render_mixin, Keyboard):
         for key in layer_area.iter_keys():
             w = key.get_border_rect().w
             histogram[w] = histogram.get(w, 0) + 1
-        most_frequent_width = max(zip(histogram.values(), histogram.keys()))[1] \
+        most_frequent_width = max(list(zip(list(histogram.values()), list(histogram.keys()))))[1] \
                               if histogram else 18
 
         # Legacy onboard had automatic tab-keys for pane switching.
@@ -460,10 +434,9 @@ class KeyboardSVG(config.kbd_render_mixin, Keyboard):
         attributes["group"]  = group
         attributes["image"]  = "close.svg"
         attributes["button"] = "true"
+        attributes["scannable"] = "false"
         self._init_key(key, attributes)
-        key.location = (rect.x, rect.y)
-        key.geometry = (rect.w, rect.h)
-        key.context.log_rect = rect
+        key.set_border_rect(rect.copy())
         keys.append(key)
 
         key = RectKey()
@@ -472,10 +445,9 @@ class KeyboardSVG(config.kbd_render_mixin, Keyboard):
         attributes["group"]  = group
         attributes["image"]  = "move.svg"
         attributes["button"] = "true"
+        attributes["scannable"] = "false"
         self._init_key(key, attributes)
-        key.location = (rect.x, rect.y)
-        key.geometry = (rect.w, rect.h)
-        key.context.log_rect = rect
+        key.set_border_rect(rect.copy())
         keys.append(key)
 
         if len(panes) > 1:
@@ -487,10 +459,7 @@ class KeyboardSVG(config.kbd_render_mixin, Keyboard):
                 attributes["label"]  = pane.id
                 attributes["button"] = "true"
                 self._init_key(key, attributes)
-                key.location = (rect.x, rect.y)
-                key.geometry = (rect.w, rect.h)
-                key.context.log_rect = rect
-
+                key.set_border_rect(rect.copy())
                 keys.append(key)
 
         layer_switch_column = LayoutBox()
@@ -504,10 +473,6 @@ class KeyboardSVG(config.kbd_render_mixin, Keyboard):
 
         return [layout]
 
-
-
-
-
     @staticmethod
     def copy_layout(src_filename, dst_filename):
         src_dir = os.path.dirname(src_filename)
@@ -525,12 +490,12 @@ class KeyboardSVG(config.kbd_render_mixin, Keyboard):
             keyboard_node = domdoc.documentElement
 
             # check layout format
-            format = 1.0
+            format = KeyboardSVG.LAYOUT_FORMAT_LEGACY
             if keyboard_node.hasAttribute("format"):
-               format = float(keyboard_node.attributes["format"].value)
+               format = Version.from_string(keyboard_node.attributes["format"].value)
             keyboard_node.attributes["id"] = dst_basename
 
-            if format < 2.0:   # layout-tree format
+            if format < KeyboardSVG.LAYOUT_FORMAT_LAYOUT_TREE:
                 raise Exceptions.LayoutFileError( \
                     _("copy_layouts failed, unsupported layout format '{}'.") \
                     .format(format))
@@ -562,7 +527,7 @@ class KeyboardSVG(config.kbd_render_mixin, Keyboard):
                 f.write(xml.encode("UTF-8"))
 
                 # copy the svg files
-                for src, dst in svg_filenames.items():
+                for src, dst in list(svg_filenames.items()):
 
                     dir, name = os.path.split(src)
                     if not dir:
@@ -598,7 +563,7 @@ class KeyboardSVG(config.kbd_render_mixin, Keyboard):
 
             layout_dir, name = os.path.split(filename)
             results = []
-            for fn in filenames.keys():
+            for fn in list(filenames.keys()):
                 dir, name = os.path.split(fn)
                 results.append(os.path.join(layout_dir, name))
 
@@ -620,7 +585,7 @@ class KeyboardSVG(config.kbd_render_mixin, Keyboard):
 
     @staticmethod
     def is_layout_node(dom_node):
-        return dom_node.tagName in [u"box", u"panel", u"key"]
+        return dom_node.tagName in ["box", "panel", "key"]
 
     @staticmethod
     def _iter_dom_nodes(dom_node):
